// White MIDI box PIC18F27Q43
// UART1 115200, 8, n, 1   UARTS 2, 3, 4, 5 31250 for MIDI
// be sure to connect any unused Rx inputs to pull-up resistors

// Default UART pin allocation                                     changed to:
//  U1 Tx      R??    ( PORTS B and C valid)  RxyPPS = 0x20          RC6
//  U1 Rx      RC7    ( PORTS B and C valid)                        (RC7)
//  U2 Tx      R??    ( PORTS B and C valid)  RxyPPS = 0x23          RC0
//  U2 Rx      RB7    ( PORTS B and C valid)                         RC1 
//  U3 Tx      R??    ( PORTS A and B valid)  RxyPPS = 0x26          RA4
//  U3 Rx      RA7    ( PORTS A and B valid)                         RA5
//  U4 Tx      R??    ( PORTS A and B valid)  RxyPPS = 0x29          RB6
//  U4 Rx      RB5    ( PORTS B and C valid)                         RB7
//  U5 Tx      R??    ( PORTS B and C valid)  RxyPPS = 0x2c          RC4
//  U5 Rx      RA5    ( PORTS A and C valid)                         RC5

// Status 28 Jan 2021
//  - need to check baud rate accuracy and remove default PPS for UART TX
//  - all five Tx and Rx pins seem OK
//  - Timer 0, INT0 (RB0), INT1 (RB1) - all seem to work OK

// PIC18F27Q43 pin use
//
//  1 MCLR
//  2 RA0 IR input
//  3 RA1
//  4 RA2
//  5 RA3 LCD R/W#
//  6 RA4 UART 3 TX
//  7 RA5 UART 3 RX
//  8 GND
//  9 OSC
// 10 OSC
// 11 RC0 UART 2 TX
// 12 RC1 UART 2 RX
// 13 RC2 UART 4 TX
// 14 RC3 UART 4 RX
//
// 15 RC4 UART 5 TX
// 16 RC5 UART 5 RX
// 17 RC6 UART 1 TX
// 18 RC7 UART 1 RX
// 19 GND
// 20 VCC
// 21 RB0 LCD D4
// 22 RB1 LCD D5
// 23 RB2 LCD D6
// 24 RB3 LCD D7
// 25 RB4 LCD RS
// 26 RB5 LCD EN
// 27 RB6 UART 4 TX     ICSP CLK
// 28 RB7 UART 4 RX     ICSP DAT

// 2x16 character LCD connections
//
//  1  GND GND
//  2  VCC +5V
//  3   Vo contrast from 10k pot
//  4   RS RB4
//  5  R/W RA3
//  6    E RB5
//  7   D0 GND
//  8   D1 GND
//  9   D2 GND
// 10   D3 GND
// 11   D4 RB0
// 12   D5 RB1
// 13   D6 RB2
// 14   D7 RB3
// 15 LED+ A      via plug-in resistor  +5V
// 16 LED- K      GND

// PICkit programming connector
// 1 Vpp   PIC pin  1 MCLR   10K to +5V
// 2 Vcc   +5V
// 3 GND   GND
// 4 PGD   PIC pin 28 RB7
// 5 PGC   PIC pin 27 RB6
// 6 AUX   nc

#define _XTAL_FREQ 64000000     // 16MHz 2-pin crystal x4 PLL for 32 MHz clock

// PIC18F27Q43 Configuration Bit Settings
// CONFIG1
#pragma config FEXTOSC = HS     // External Oscillator Selection (HS (crystal oscillator) above 8 MHz)
#pragma config RSTOSC = EXTOSC_4PLL// Reset Oscillator Selection (EXTOSC with 4x PLL, with EXTOSC operating per FEXTOSC bits)
// CONFIG2
#pragma config CLKOUTEN = OFF   // Clock out Enable bit (CLKOUT function is disabled)
#pragma config PR1WAY = OFF     // PRLOCKED One-Way Set Enable bit (PRLOCKED bit can be set and cleared repeatedly)
#pragma config CSWEN = OFF      // Clock Switch Enable bit (The NOSC and NDIV bits cannot be changed by user software)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable bit (Fail-Safe Clock Monitor disabled)
// CONFIG3
#pragma config MCLRE = EXTMCLR  // MCLR Enable bit (If LVP = 0, MCLR pin is MCLR; If LVP = 1, RE3 pin function is MCLR )
#pragma config PWRTS = PWRT_64  // Power-up timer selection bits (PWRT set at 64ms)
#pragma config MVECEN = ON      // Multi-vector enable bit (Multi-vector enabled, Vector table used for interrupts)
#pragma config IVT1WAY = OFF    // IVTLOCK bit One-way set enable bit (IVTLOCKED bit can be cleared and set repeatedly)
#pragma config LPBOREN = OFF    // Low Power BOR Enable bit (Low-Power BOR disabled)
#pragma config BOREN = SBORDIS  // Brown-out Reset Enable bits (Brown-out Reset enabled , SBOREN bit is ignored)
// CONFIG4
#pragma config BORV = VBOR_1P9  // Brown-out Reset Voltage Selection bits (Brown-out Reset Voltage (VBOR) set to 1.9V)
#pragma config ZCD = OFF        // ZCD Disable bit (ZCD module is disabled. ZCD can be enabled by setting the ZCDSEN bit of ZCDCON)
#pragma config PPS1WAY = OFF    // PPSLOCK bit One-Way Set Enable bit (PPSLOCKED bit can be set and cleared repeatedly (subject to the unlock sequence))
#pragma config STVREN = ON      // Stack Full/Underflow Reset Enable bit (Stack full/underflow will cause Reset)
#pragma config LVP = OFF        // Low Voltage Programming Enable bit (HV on MCLR/VPP must be used for programming)
#pragma config XINST = OFF      // Extended Instruction Set Enable bit (Extended Instruction Set and Indexed Addressing Mode disabled)
// CONFIG5
#pragma config WDTCPS = WDTCPS_31// WDT Period selection bits (Divider ratio 1:65536; software control of WDTPS)
#pragma config WDTE = OFF       // WDT operating mode (WDT Disabled; SWDTEN is ignored)
// CONFIG6
#pragma config WDTCWS = WDTCWS_7// WDT Window Select bits (window always open (100%); software control; keyed access not required)
#pragma config WDTCCS = SC      // WDT input clock selector (Software Control)
// CONFIG7
#pragma config BBSIZE = BBSIZE_512// Boot Block Size selection bits (Boot Block size is 512 words)
#pragma config BBEN = OFF       // Boot Block enable bit (Boot block disabled)
#pragma config SAFEN = OFF      // Storage Area Flash enable bit (SAF disabled)
#pragma config DEBUG = OFF      // Background Debugger (Background Debugger disabled)
// CONFIG8
#pragma config WRTB = OFF       // Boot Block Write Protection bit (Boot Block not Write protected)
#pragma config WRTC = OFF       // Configuration Register Write Protection bit (Configuration registers not Write protected)
#pragma config WRTD = OFF       // Data EEPROM Write Protection bit (Data EEPROM not Write protected)
#pragma config WRTSAF = OFF     // SAF Write protection bit (SAF not Write Protected)
#pragma config WRTAPP = OFF     // Application Block write protection bit (Application Block not write protected)
// CONFIG10
#pragma config CP = OFF         // PFM and Data EEPROM Code Protection bit (PFM and Data EEPROM code protection disabled)

#include <xc.h>                 // always needed 
#include <stdio.h>              // allows fprint library function (among others))
#include <stdlib.h>             // not sure if this is really needed, but left just in case

// global variables
volatile unsigned char u1_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u1_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u1_rx_buf[257];                                                   // circular buffer for UART 1 Rx - must be 256 bytes (I added one spare)
volatile unsigned char u2_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u2_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u2_rx_buf[257];                                                   // circular buffer for UART 2 Rx - must be 256 bytes (I added one spare)
volatile unsigned char u3_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u3_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u3_rx_buf[257];                                                   // circular buffer for UART 3 Rx - must be 256 bytes (I added one spare)
volatile unsigned char u4_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u4_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u4_rx_buf[257];                                                   // circular buffer for UART 4 Rx - must be 256 bytes (I added one spare)
volatile unsigned char u5_rx_in_ptr;                                            // add byte to circular buffer       - must be 8 bits for wrap-around
volatile unsigned char u5_rx_out_ptr;                                           // remove byte from circular buffer  - must be 8 bits for wrap-around
unsigned char u5_rx_buf[257];                                                   // circular buffer for UART 5 Rx - must be 256 bytes (I added one spare)
volatile unsigned long count10ms;                                               // count tens of milliseconds

// UART 1 receive interrupt
void __interrupt(irq(U1RX),high_priority) u1Int(void){
    U1RXIF = 0;                                                                 // clear interrupt condition
    u1_rx_buf[u1_rx_in_ptr] = U1RXB;                                            // put character into buffer
    u1_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// UART 2 receive interrupt
void __interrupt(irq(U2RX),high_priority) u2Int(void){
    U2RXIF = 0;                                                                 // clear interrupt condition
    u2_rx_buf[u2_rx_in_ptr] = U2RXB;                                            // put character into buffer
    u2_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// UART 3 receive interrupt
void __interrupt(irq(U3RX),high_priority) u3Int(void){
    U3RXIF = 0;                                                                 // clear interrupt condition
    u3_rx_buf[u3_rx_in_ptr] = U3RXB;                                            // put character into buffer
    u3_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// UART 4 receive interrupt
void __interrupt(irq(U4RX),high_priority) u4Int(void){
    U4RXIF = 0;                                                                 // clear interrupt condition
    u4_rx_buf[u4_rx_in_ptr] = U4RXB;                                            // put character into buffer
    u4_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// UART 5 receive interrupt
void __interrupt(irq(U5RX),high_priority) u5Int(void){
    U5RXIF = 0;                                                                 // clear interrupt condition
    u5_rx_buf[u5_rx_in_ptr] = U5RXB;                                            // put character into buffer
    u5_rx_in_ptr++;                                                             // update buffer pointer
    return;
}

// INT 0 Interrupt
void __interrupt(irq(INT0),high_priority) i0Int(void){
   
    INT0IF = 0;                                                             // clear interrupt condition
}

// INT 1 Interrupt
void __interrupt(irq(INT1),high_priority) i1Int(void){
    INT1IF = 0;                                                             // clear interrupt condition
}

// Timer 0 Interrupt every 10 ms
void __interrupt(irq(TMR0),high_priority) T0Int(void){
    TMR0IF = 0;                                                             // clear interrupt
    TMR0H = 0x63;                                                           // adjust for 10ms interrupt
    TMR0L = 0xc0;                                                           // adjust for 10ms interrupt
    count10ms ++;
}

//default - catch-all interrupt handler - should never reach here
void __interrupt(irq(default),high_priority) deflt(void){
    LATA = 0xff;                                                                // light all LEDs to show something bad happened
    while(1);                                                                   // infinite loop for aiding debugging
    return;
}

// initialise interrupts for peripherals
void Interrupt_Init(void){
    T0CON0 = 0x90;                                                              // T0 on, 16 bit
    T0CON1 = 0x40;                                                              // fosc/4
    TMR0H = 0x63;                                                               // adjust for 10ms interrupt
    TMR0L = 0xc0;                                                               // adjust for 10ms interrupt
    TMR0IE = 1;                                                                 // enable timer 0 interrupts
    
    INT0IF = 0;                                                                 // clear INT0 interrupt
    INT0IE = 1;                                                                 // enable INT0 on RB0 falling edge
    INT1IF = 0;                                                                 // clear INT1 interrupt
    INT1IE = 1;                                                                 // enable INT1 on RB1 falling edge

}

// blocking function to send byte to serial port - required for fprintf function and similar
void putch(unsigned char send_char){                     
    
    while(! U1TXIF);                                                            // wait Tx ready
    U1TXB = send_char;                                                          // send the byte
    
    while(! U2TXIF);                                                            // wait Tx ready
    U2TXB = send_char;                                                          // send the byte
    
    while(! U3TXIF);                                                            // wait Tx ready
    U3TXB = send_char;                                                          // send the byte
    
    while(! U4TXIF);                                                            // wait Tx ready
    U4TXB = send_char;                                                          // send the byte
    
    while(! U5TXIF);                                                            // wait Tx ready
    U5TXB = send_char;                                                          // send the byte
}

// initialise desired default UARTs - could remove some Rx PPS allocations where default is used, but did not
void UART_Init(void){
    
    u1_rx_in_ptr = 0;                                                           // reset buffer pointer
    u1_rx_out_ptr = 0;                                                          // reset buffer pointer
    u2_rx_in_ptr = 0;                                                           // reset buffer pointer
    u2_rx_out_ptr = 0;                                                          // reset buffer pointer
    u3_rx_in_ptr = 0;                                                           // reset buffer pointer
    u3_rx_out_ptr = 0;                                                          // reset buffer pointer
    u4_rx_in_ptr = 0;                                                           // reset buffer pointer
    u4_rx_out_ptr = 0;                                                          // reset buffer pointer
    u5_rx_in_ptr = 0;                                                           // reset buffer pointer
    u5_rx_out_ptr = 0;                                                          // reset buffer pointer

    TRISCbits.TRISC6 = 0;                                                       // set RC6 as output for UART 1 Tx
    TRISCbits.TRISC7 = 1;                                                       // set RC7 as input for UART 1 Rx
    U1CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U1BRG = 0x8a;                                                               // desired baud rate (0x8a for 115200 baud))
    RC6PPS = 0x20;                                                              // UART 1 Tx on RC6 (default) (ports B and C valid)
    U1RXPPS = 0x17;                                                             // UART 1 Rx on RC7 (default) (ports B and C valid)
    U1CON1 = 0x80;                                                              // serial port enable
    
    TRISCbits.TRISC0 = 0;                                                       // set RC0 as output for UART 2 Tx
    TRISCbits.TRISC1 = 1;                                                       // set RC1 as input for UART 2 Rx
    U2CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U2BRG = 0x1fd;                                                              // desired baud rate (0x1fd for 31250 baud)
    RC0PPS = 0x23;                                                              // UART 2 Tx on RC0 (ports B and C valid)
    U2RXPPS = 0x11;                                                             // UART 2 Rx on RC1 (ports B and C valid)
    U2CON1 = 0x80;                                                              // serial port enable

    TRISAbits.TRISA4 = 0;                                                       // set RA4 as output for UART 3 Tx
    TRISAbits.TRISA5 = 1;                                                       // set RA5 as input for UART 3 Rx
    U3CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U3BRG = 0x1fd;                                                              // desired baud rate (0x1fd for 31250 baud)
    RA4PPS = 0x26;                                                              // UART 3 Tx on RA4 (ports A and B valid)
    U3RXPPS = 0x05;                                                             // UART 3 Rx on RA5 (ports A and B valid)
    U3CON1 = 0x80;                                                              // serial port enable
    RA7PPS = 0;                                                                 // restore old UART3 Rx default to GPIO

    TRISBbits.TRISB6 = 0;                                                       // set RB6 as output for UART 4 Tx
    TRISBbits.TRISB7 = 1;                                                       // set RB7 as input for UART 4 Rx
    U4CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U4BRG = 0x1fd;                                                              // desired baud rate (0x1fd for 31250 baud)
    RB6PPS = 0x29;                                                              // UART 4 Tx on RB6 (ports A and B valid)
    U4RXPPS = 0x0f;                                                             // UART 4 Rx on RB7 (ports B and C valid)
    U4CON1 = 0x80;                                                              // serial port enable

    TRISCbits.TRISC4 = 0;                                                       // set RC4 as output for UART 5 Tx
    TRISCbits.TRISC5 = 1;                                                       // set RC5 as input for UART 5 Rx
    U5CON0 = 0xb0;                                                              // high speed, tx enab, rx, enab, 8 bit                     
    U5BRG = 0x1fd;                                                              // desired baud rate (0x1fd for 31250 baud)
    RC4PPS = 0x2c;                                                              // UART 5 Tx on RC4 (ports B and C valid)
    U5RXPPS = 0x15;                                                             // UART 5 Rx on RC5 (ports A and C valid)
    U5CON1 = 0x80;                                                              // serial port enable
    RA5PPS = 0;                                                                 // restore old UART5 Rx default to GPIO
    
    __delay_ms(100);                                                            // UART settle time
    
    U1RXIE = 1;                                                                 // enable UART 1 Rx interrupt
    U2RXIE = 1;                                                                 // enable UART 2 Rx interrupt    
    U3RXIE = 1;                                                                 // enable UART 3 Rx interrupt    
    U4RXIE = 1;                                                                 // enable UART 4 Rx interrupt    
    U5RXIE = 1;                                                                 // enable UART 5 Rx interrupt    
}

// set all ports output to start with - will be modified later
void IO_Init(void){
    LATA = 0;                                                                   // all outputs low
    LATB = 0;                                                                   // all outputs low
    LATC = 0;                                                                   // all outputs low
    ANSELA = 0;                                                                 // all digital
    ANSELB = 0;                                                                 // all digital
    ANSELC = 0;                                                                 // all digital
    TRISA = 0;                                                                  // all output for diagnostic LEDs
}

int main(int argc, char** argv) {
  unsigned char u1_temp;
  unsigned char u2_temp;
  unsigned char u3_temp;
  unsigned char u4_temp;
  unsigned char u5_temp;

    IO_Init();                                                                  // set up ports as desired
    Interrupt_Init();
    UART_Init();                                                                // initialise desired UART as default
    GIE = 1;                                                                    // all interrupts enabled
    
    printf("\r\n\nMIDItest 29 Jan 2021\r\n\n ");
    
    while(1){                                                                   // main loop
        if(u1_rx_in_ptr != u1_rx_out_ptr){                                      // is there any data in buffer?
            u1_temp = u1_rx_buf[u1_rx_out_ptr];                                 //  ... if so, then get a byte
            u1_rx_out_ptr++;                                                    // remove byte from buffer
            putch(u1_temp);                                                     // echo the byte
        }
        if(u2_rx_in_ptr != u2_rx_out_ptr){                                      // is there any data in buffer?
            u2_temp = u2_rx_buf[u2_rx_out_ptr];                                 //  ... if so, then get a byte
            u2_rx_out_ptr++;                                                    // remove byte from buffer
            putch(u2_temp);                                                     // echo the byte
        }
        if(u3_rx_in_ptr != u3_rx_out_ptr){                                      // is there any data in buffer?
            u3_temp = u3_rx_buf[u3_rx_out_ptr];                                 //  ... if so, then get a byte
            u3_rx_out_ptr++;                                                    // remove byte from buffer
            putch(u3_temp);                                                     // echo the byte
        }
        if(u4_rx_in_ptr != u4_rx_out_ptr){                                      // is there any data in buffer?
            u4_temp = u4_rx_buf[u4_rx_out_ptr];                                 //  ... if so, then get a byte
            u4_rx_out_ptr++;                                                    // remove byte from buffer
            putch(u4_temp);                                                     // echo the byte
        }
        if(u5_rx_in_ptr != u5_rx_out_ptr){                                      // is there any data in buffer?
            u5_temp = u5_rx_buf[u5_rx_out_ptr];                                 //  ... if so, then get a byte
            u5_rx_out_ptr++;                                                    // remove byte from buffer
            putch(u5_temp);                                                     // echo the byte
        }
    
        LATAbits.LATA5 = (unsigned char)count10ms >> 6;                         // flash LED to show timer T0 is working
        putch(0x55);
        __delay_ms(1000);
    }
}

